<?php

namespace Drupal\flysystem\Adapter;

use League\Flysystem\Config;
use League\Flysystem\FileAttributes;
use League\Flysystem\FilesystemAdapter;

/**
 * A Flysystem adapter implementing caching with Drupal's Cache API.
 */
class DrupalCacheAdapter implements FilesystemAdapter {

  /**
   * The Flysystem adapter to cache data for.
   *
   * @var \League\Flysystem\FilesystemAdapter
   */
  protected FilesystemAdapter $adapter;

  /**
   * The cache backend to store data in.
   *
   * @var \Drupal\flysystem\Adapter\CacheItemBackend
   */
  protected CacheItemBackend $cacheItemBackend;

  /**
   * The scheme of the stream wrapper used for this adapter.
   *
   * @var string
   */
  protected string $scheme;

  /**
   * Constructs a new caching Flysystem adapter.
   *
   * @param string $scheme
   *   The scheme of the stream wrapper used for this adapter.
   * @param \League\Flysystem\FilesystemAdapter $adapter
   *   The flysystem adapter to cache data for.
   * @param \Drupal\flysystem\Adapter\CacheItemBackend $cacheItemBackend
   *   The cache backend to store data in.
   */
  public function __construct(string $scheme, FilesystemAdapter $adapter, CacheItemBackend $cacheItemBackend) {
    $this->scheme = $scheme;
    $this->adapter = $adapter;
    $this->cacheItemBackend = $cacheItemBackend;
  }

  /**
   * {@inheritdoc}
   */
  public function write($path, $contents, Config $config): void {
    $this->adapter->write($path, $contents, $config);
  }

  /**
   * {@inheritdoc}
   */
  public function writeStream($path, $contents, Config $config): void {
    $this->adapter->writeStream($path, $contents, $config);
  }

  /**
   * {@inheritdoc}
   */
  public function copy($source, $destination, $config): void {
    $this->adapter->copy($source, $destination, $config);

    $item = $this->cacheItemBackend->load($source);
    $newitem = clone $item;
    $this->cacheItemBackend->set($destination, $newitem);
  }

  /**
   * {@inheritdoc}
   */
  public function delete($path): void {
    $this->adapter->delete($path);

    $this->cacheItemBackend->delete($path);
  }

  /**
   * {@inheritdoc}
   */
  public function setVisibility($path, $visibility): void {
    $this->adapter->setVisibility($path, $visibility);

    $this->updateMetadata($path, ['visibility' => $visibility]);
  }

  /**
   * {@inheritdoc}
   */
  public function read($path): string {
    return $this->adapter->read($path);
  }

  /**
   * {@inheritdoc}
   */
  public function readStream($path) {
    return $this->adapter->readStream($path);
  }

  /**
   * {@inheritdoc}
   */
  public function listContents($path = '', $deep = FALSE): iterable {
    // Don't cache directory listings to avoid having to keep track of
    // incomplete cache entries.
    // @todo This could be a good place for a microcache?
    return $this->adapter->listContents($path, $deep);
  }

  /**
   * Something something.
   *
   * @param string $path
   *   The file path.
   *
   * @return array
   *   The metadata.
   */
  public function getMetadata(string $path): array {
    $item = $this->cacheItemBackend->load($path);

    if ($metadata = $item->getMetadata()) {
      return $metadata;
    }

    $filesize = $this->adapter->fileSize($path);
    $lastmodified = $this->adapter->lastModified($path);
    $mimetype = $this->adapter->mimeType($path);
    $visibility = $this->adapter->visibility($path);

    // $metadata = $this->adapter->getMetadata($path);
    $item->updateMetadata([
      'fileSize' => $filesize,
      'lastModified' => $lastmodified,
      'mimeType' => $mimetype,
      'visibility' => $visibility,
    ]);
    $this->cacheItemBackend->set($path, $item);

    return $metadata;
  }

  /**
   * Fetches a specific key from metadata.
   *
   * @param string $path
   *   The path to load metadata for.
   * @param string $key
   *   The key in metadata, such as 'mimetype', to load metadata for.
   *
   * @return string|bool
   *   The array of metadata.
   */
  protected function fetchMetadataKey(string $path, string $key): string|bool {
    $item = $this->cacheItemBackend->load($path);

    if (($metadata = $item->getMetadata()) && isset($metadata[$key])) {
      return $metadata[$key];
    }

    $metadata = $this->adapter->$key($path);
    $updatedMetadata = $this->updateMetadata($path, [$key => $metadata->$key()]);
    return $updatedMetadata[$key] ?? FALSE;
  }

  /**
   * Updates the metadata for a given path.
   *
   * @param string $path
   *   The path of file file or directory.
   * @param array|false $metadata
   *   The metadata to update.
   *
   * @return array|false
   *   Returns the value passed in as metadata.
   */
  protected function updateMetadata(string $path, array|false $metadata): array|false {
    if (!empty($metadata)) {
      $item = $this->cacheItemBackend->load($path);
      $item->updateMetadata($metadata);
      $this->cacheItemBackend->set($path, $item);
    }

    return $metadata;
  }

  /**
   * {@inheritdoc}
   */
  public function fileExists(string $path): bool {
    if ($this->cacheItemBackend->has($path)) {
      return TRUE;
    }

    // Always check the upstream adapter for new files.
    // @todo This could be a good place for a microcache?
    return $this->adapter->fileExists($path);
  }

  /**
   * {@inheritdoc}
   */
  public function directoryExists(string $path): bool {
    if ($this->cacheItemBackend->has($path)) {
      return TRUE;
    }

    // Always check the upstream adapter for new directoies.
    // @todo This could be a good place for a microcache?
    return $this->adapter->directoryExists($path);
  }

  /**
   * {@inheritdoc}
   */
  public function deleteDirectory(string $path): void {
    // Before the delete we need to know what files are in the directory.
    $contents = $this->adapter->listContents($path, TRUE);

    $this->adapter->deleteDirectory($path);

    $paths = array_column($contents, 'path');
    $this->cacheItemBackend->deleteMultiple($paths);
  }

  /**
   * {@inheritdoc}
   */
  public function createDirectory(string $path, Config $config): void {
    $this->adapter->createDirectory($path, $config);

    // Warm the metadata cache.
    $item = new CacheItem();
    $this->cacheItemBackend->set($path, $item);
  }

  /**
   * {@inheritdoc}
   */
  public function visibility(string $path): FileAttributes {
    $visibility = $this->fetchMetadataKey($path, 'visibility');
    return new FileAttributes($path, NULL, $visibility);
  }

  /**
   * {@inheritdoc}
   */
  public function mimeType(string $path): FileAttributes {
    $mimeType = $this->fetchMetadataKey($path, 'mimeType');
    return new FileAttributes($path, NULL, NULL, NULL, $mimeType);
  }

  /**
   * {@inheritdoc}
   */
  public function lastModified(string $path): FileAttributes {
    return $this->adapter->lastModified($path);
    // Return new FileAttributes($path, NULL, NULL, $lastModified);.
  }

  /**
   * {@inheritdoc}
   */
  public function fileSize(string $path): FileAttributes {
    return $this->adapter->fileSize($path);
  }

  /**
   * {@inheritdoc}
   */
  public function move(string $source, string $destination, Config $config): void {
    $this->adapter->move($source, $destination, $config);

    $item = $this->cacheItemBackend->load($source);
    $newitem = clone $item;
    $this->cacheItemBackend->set($destination, $newitem);
    $this->cacheItemBackend->delete($source);
  }

}
