<?php

namespace Drupal\flysystem;

use Drupal\Core\StreamWrapper\StreamWrapperInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use League\Flysystem\Filesystem;

/**
 * FlysystemStreamWrapper implementation for Drupal compatibility.
 *
 * Was previously FlysystemBridge.
 */
class FlysystemBridge extends Filesystem {

  use StringTranslationTrait;

  /**
   * The registered filesystems.
   *
   * @var \League\Flysystem\FilesystemOperator[]
   */
  protected static $filesystems = [];

  /**
   * Optional configuration.
   *
   * @var array
   */
  protected static $config = [];

  /**
   * The default configuration.
   *
   * @var array
   */
  protected static $defaultConfiguration = [
    'permissions' => [
      'dir' => [
        'private' => 0700,
        'public' => 0755,
      ],
      'file' => [
        'private' => 0600,
        'public' => 0644,
      ],
    ],
    'metadata' => ['timestamp', 'size', 'visibility'],
    'public_mask' => 0044,
  ];

  /**
   * Scheme protocol.
   */
  private $uri;

  /**
   * {@inheritdoc}
   */
  public static function getType() {
    return StreamWrapperInterface::WRITE_VISIBLE;
  }

  /**
   * {@inheritdoc}
   */
  public function getName() {
    $scheme = $this->getProtocol();
    $name = $this->getFactory()->getSettings($scheme)['name'];
    $default = $this->t('Flysystem: @scheme', ['@scheme' => $scheme]);

    return $name !== '' ? $name : $default;
  }

  /**
   * {@inheritdoc}
   */
  public function getDescription() {
    $scheme = $this->getProtocol();
    $description = $this->getFactory()->getSettings($scheme)['description'];
    $default = $this->t('Flysystem: @scheme', ['@scheme' => $scheme]);

    return $description !== '' ? $description : $default;
  }

  /**
   * {@inheritdoc}
   */
  public function getUri() {
    return $this->uri;
  }

  /**
   * {@inheritdoc}
   */
  public function setUri($uri) {
    $this->uri = $uri;
  }

  /**
   * {@inheritdoc}
   */
  public function getExternalUrl() {
    return $this
      ->getFactory()
      ->getPlugin($this->getProtocol())
      ->getExternalUrl($this->uri);
  }

  /**
   * {@inheritdoc}
   */
  public function realpath() {
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function dirname($uri = NULL) {
    if (!isset($uri)) {
      $uri = $this->uri;
    }

    [$scheme, $target] = explode('://', $uri, 2);

    return $scheme . '://' . ltrim(($target === '.' ? '' : $target), '\/');
  }

  /**
   * Returns the filesystem for a given scheme.
   *
   * @param string $scheme
   *   The scheme.
   *
   * @return \Drupal\flysystem\Flysystem\Interface\FilesystemInterface
   *   The filesystem for the scheme.
   */
  protected function getFilesystemForScheme($scheme) {
    if (!isset(static::$filesystems[$scheme])) {
      static::$filesystems[$scheme] = $this->getFactory()->getFilesystem($scheme);
      static::$config[$scheme] = static::$defaultConfiguration;
      static::$config[$scheme]['permissions']['dir']['public'] = 0777;
      // This is now an adapter.
      static::registerPlugins($scheme, static::$filesystems[$scheme]);
    }

    return static::$filesystems[$scheme];
  }

  /**
   * {@inheritdoc}
   */
  protected function getFilesystem() {
    if (!isset($this->filesystem)) {
      $this->filesystem = $this->getFilesystemForScheme($this->getProtocol());
    }

    return $this->filesystem;
  }

  /**
   * Returns the filesystem factory.
   *
   * @return \Drupal\flysystem\FlysystemStreamWrapperManager
   *   The Flysystem factory.
   */
  protected function getFactory() {
    return \Drupal::service('flysystem_factory');
  }

  /**
   * Returns the protocol from the internal URI.
   *
   * @return string The protocol.
   */
  protected function getProtocol() {
    return substr($this->uri, 0, strpos($this->uri, '://'));
  }

  /**
   * Registers adapters on the filesystem.
   *
   * @param string $protocol
   * @param \League\Flysystem\Filesystem $filesystem
   */
  protected static function registerAdapters($protocol, Filesystem $filesystem) {
    $filesystem->addPlugin($stat);
  }

}
