<?php

namespace Drupal\flysystem;

use Drupal\Component\Utility\Unicode;
use Drupal\Core\File\Exception\FileException;
use Drupal\Core\File\Exception\FileWriteException;
use Drupal\Core\File\Exception\InvalidStreamWrapperException;
use Drupal\Core\File\Exception\NotRegularFileException;
use Drupal\Core\File\FileSystem;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Site\Settings;
use Drupal\Core\StreamWrapper\StreamWrapperInterface;
use Drupal\Core\StreamWrapper\StreamWrapperManager;
use Drupal\Core\StreamWrapper\StreamWrapperManagerInterface;
use League\Flysystem\FilesystemAdapter;

/**
 *
 */
class FlysystemDrupalFileSystem extends FileSystem {

  /**
   * The inner service.
   *
   * @var \Drupal\Core\File\FileSystem
   */
  protected $decoratedService;

  /**
   * The site settings.
   *
   * @var \Drupal\Core\Site\Settings
   */
  protected $settings;

  /**
   * The stream wrapper manager.
   *
   * @var \Drupal\Core\StreamWrapper\StreamWrapperManagerInterface
   */
  protected $streamWrapperManager;

  /**
   * Constructs a new FileSystem.
   *
   * Leverages Drupal's existing FileSystem object, enhanced by Flysystem's
   * Adapters.
   *
   * Flysystem Adapters are implemented as part of instantiating a Flysystem
   * Operator.  In this module, Flysystem Operators are registered using
   * Drupal's StreamWrapperManager.
   *
   * @param \Drupal\Core\File\FileSystem $decorated_service
   *   Drupal Core's FileSystem service, which is being decorated here.
   * @param \Drupal\Core\StreamWrapper\StreamWrapperManagerInterface $stream_wrapper_manager
   *   The stream wrapper manager.
   * @param \Drupal\Core\Site\Settings $settings
   *   The site settings.
   */
  public function __construct(FileSystem $decorated_service, StreamWrapperManagerInterface $stream_wrapper_manager, Settings $settings) {
    $this->decoratedService = $decorated_service;
    $this->streamWrapperManager = $stream_wrapper_manager;
    $this->settings = $settings;
  }

  /**
   * {@inheritdoc}
   *
   * @todo rewrite to leverage Flysystem (yes)
   * @see \Drupal\Core\File\FileSystem::moveUploadedFile()
   */
  public function moveUploadedFile($filename, $uri) {
    return $this->decoratedService->moveUploadedFile($filename, $uri);
  }

  /**
   * {@inheritdoc}
   *
   * @todo finish writing, see inline todo comments.
   */
  public function chmod($uri, $mode = NULL) {
    if (!isset($mode)) {
      /** @var \Drupal\Core\StreamWrapper\StreamWrapperInterface $wrapper */
      $wrapper = $this->streamWrapperManager->getViaUri($uri);
      if ($wrapper instanceof FilesystemAdapter) {
        try {
          $drupalMode = NULL;
          if ($wrapper->directoryExists($uri)) {
            // @todo May have to write a Flysystem VisibilityConverter for
            // Local file adapter that substitutes Drupal file and directory
            // permission settings for the default settings in Flysystem.
            $drupalMode = $this->settings->get('file_chmod_directory', static::CHMOD_DIRECTORY);
          }
          elseif ($wrapper->fileExists($uri)) {
            $drupalMode = $this->settings->get('file_chmod_file', static::CHMOD_FILE);
          }
          return $wrapper->setVisibility($uri, $drupalMode);
        }
        catch (\Exception $e) {
          // @todo write exception handler for possible exceptions thrown
          // by FlysystemAdapter::setVisibility();
        }
      }
    }
    return $this->decoratedService->chmod($uri, $mode);
  }

  /**
   * {@inheritdoc}
   *
   * @todo finish writing, see inline todo comments.
   */
  public function mkdir($uri, $mode = NULL, $recursive = FALSE, $context = NULL) {
    /** @var \Drupal\Core\StreamWrapper\StreamWrapperInterface $wrapper */
    $wrapper = $this->streamWrapperManager->getViaUri($uri);
    if ($wrapper instanceof FilesystemAdapter) {
      try {
        // @todo figure out how to utilize directory permissions, see notes on
        // Flysystem VisibilityConverter.
        // @see \League\Flysystem\Filesystem::createDirectory()
        $wrapper->createDirectory($uri);
        return TRUE;
      }
      catch (\Exception $e) {
        // @todo write exception handler to handle Flysystem Exception, and
        // return a proper return code, which is FALSE.
      }
    }
    return $this->decoratedService->mkdir($uri, $mode, $recursive, $context);
  }

  /**
   * {@inheritdoc}
   *
   * @todo finish writing, see inline todo comments.
   */
  public function rmdir($uri, $context = NULL) {
    /** @var \Drupal\Core\StreamWrapper\StreamWrapperInterface $wrapper */
    $wrapper = $this->streamWrapperManager->getViaUri($uri);
    if ($wrapper instanceof FilesystemAdapter) {
      try {
        // @todo figure out how to utilize directory permissions, see notes on
        // Flysystem VisibilityConverter.
        // @see \League\Flysystem\Filesystem::deleteDirectory()
        $wrapper->deleteDirectory($uri);
        return TRUE;
      }
      catch (\Exception $e) {
        // @todo write exception handler to handle Flysystem Exception, and
        // return a proper return code, which is FALSE.
      }
    }
    return $this->decoratedService->rmdir($uri, $context);
  }

  /**
   * {@inheritdoc}
   *
   * @todo Rewrite to leverage Flysystem (yes)
   */
  public function copy($source, $destination, $replace = self::EXISTS_RENAME) {
    /** @var \Drupal\Core\StreamWrapper\StreamWrapperInterface $wrapper */
    $wrapper = $this->streamWrapperManager->getViaScheme(StreamWrapperManager::getScheme($source));
    if ($wrapper instanceof FilesystemAdapter) {
      try {
        // @todo figure out how to utilize directory permissions, see notes on
        // Flysystem VisibilityConverter.
        // @see \League\Flysystem\Filesystem::copy().
        $wrapper->copy($source, $destination);
        return $destination;
      }
      catch (\Exception $e) {
        // @todo write exception handler to handle Flysystem Exception, and
        // return a proper return code, which is FALSE.
      }
    }
    return $this->decoratedService->copy($source, $destination, $replace);
  }

  /**
   * {@inheritdoc}
   *
   * @todo finish writing, see inline todo comments.
   */
  public function delete($path) {
    /** @var \Drupal\Core\StreamWrapper\StreamWrapperManager $wrappe */
    $wrapper = $this->streamWrapperManager->getViaUri($path);
    if ($wrapper instanceof FilesystemAdapter) {
      if ($wrapper->fileExists($path) || $wrapper->directoryExists($path)) {
        if ($wrapper->directoryExists($path)) {
          throw new NotRegularFileException("Cannot delete '$path' because it is a directory. Use deleteRecursive() instead.");
        }
        if ($wrapper->fileExists($path)) {
          try {
            $wrapper->delete($path);
            return TRUE;
          }
          catch (\Exception $e) {
            // @todo modify \Exception to use class from Flysystem,
            // and throw a proper exception.
          }
        }
        // If file does not exist, return TRUE, that is the intended result.
        return TRUE;
      }
    }    
    return $this->decoratedService->delete($path);
  }

  /**
   * {@inheritdoc}
   */
  public function deleteRecursive($path, callable $callback = NULL) {
    if ($callback) {
      return $this->decoratedService->deleteRecursive($path, $callback);
    }
    /** @var \Drupal\Core\StreamWrapper\StreamWrapperInterface $wrapper */
    $wrapper = $this->streamWrapperManager->getViaUri($path);
    if ($wrapper instanceof FilesystemAdapter) {
      if (!$wrapper->fileExists($path)) {
        return TRUE;
      }

      if ($wrapper->directoryExists($path)) {
        $dir = $wrapper->listContents($path);
        foreach ($dir as $entry) {
          $entry_path = $path . '/' . $entry;
          $this->deleteRecursive($entry_path, $callback);
        }
        return $this->rmdir($path);
      }
      return $this->delete($path);
    }
    return $this->decoratedService->deleteRecursive($path, NULL);
  }

  /**
   * {@inheritdoc}
   *
   * @todo finish writing, see inline todo comments.
   */
  public function move($source, $destination, $replace = self::EXISTS_RENAME) {
    /** @var \Drupal\Core\StreamWrapper\StreamWrapperInterface $wrapper */
    $wrapper = $this->streamWrapperManager->getViaScheme(StreamWrapperManager::getScheme($source));
    if ($wrapper instanceof FilesystemAdapter) {
      try {
        // @todo figure out how to utilize directory permissions, see notes on
        // Flysystem VisibilityConverter.
        // @see \League\Flysystem\Filesystem::move().
        // What do we do with the $replace flag?
        $wrapper->move($source, $destination);
        return $destination;
      }
      catch (\Exception $e) {
        // @todo write exception handler to handle Flysystem Exception, and
        // return a proper return code, which is FALSE.
      }
    }
    return $this->decoratedService->move($source, $destination, $replace);
  }

  /**
   * {@inheritdoc}
   *
   * @todo finish writing, see inline todo comments.
   */
  public function saveData($data, $destination, $replace = self::EXISTS_RENAME) {
    // Write the data to a temporary file.
    // The assumption here is that temporary files will always use the local
    // filespace.
    $temp_name = $this->tempnam('temporary://', 'file');
    if (file_put_contents($temp_name, $data) === FALSE) {
      throw new FileWriteException("Temporary file '$temp_name' could not be created.");
    }
    if (!$this->streamWrapperManager->isValidUri($destination)) {
      throw new InvalidStreamWrapperException("Invalid stream wrapper: {$destination}");
    }
    /** @var \Drupal\Core\StreamWrapper\StreamWrapperInterface $wrapper */
    $wrapper = $this->streamWrapperManager->getViaScheme(StreamWrapperManager::getScheme($destination));
    if ($wrapper instanceof FilesystemAdapter) {
      // Move the file to its final destination.
      try {
        // @todo figure out how to utilize directory permissions, see notes on
        // Flysystem VisibilityConverter.
        // @see \League\Flysystem\Filesystem::move().
        // Do we need to use the $replace flag here in the FlysystemAdapter?
        $wrapper->move($temp_name, $destination);
        return $destination;
      }
      catch (\Exception $e) {
        // @todo write exception handler to handle Flysystem Exception, and
        // return a proper return code, which is FALSE.
      }
    }
    return $this->decoratedService->move($temp_name, $destination, $replace);
  }

  /**
   * {@inheritdoc}
   *
   * @todo finish writing, see inline todo comments.
   * @see \League\Flysystem\FilesystemAdapter::directoryExists()
   */
  public function prepareDirectory(&$directory, $options = self::MODIFY_PERMISSIONS) {
    if (!$this->streamWrapperManager->isValidUri($directory)) {
      // Only trim if we're not dealing with a stream.
      $directory = rtrim($directory, '/\\');
    }
    /** @var \Drupal\Core\StreamWrapper\StreamWrapperInterface $wrapper */
    $wrapper = $this->streamWrapperManager->getViaScheme(StreamWrapperManager::getScheme($directory));
    if ($wrapper instanceof FilesystemAdapter) {
      // How do we use the flags, self::MODIFY_PERMISSIONS and static::CREATE_DIRECTORY here?
      // @see Drupal\Core\File\FileSystemInterface::prepareDirectory().
      try {
        // Let mkdir() recursively create directories and use the default
        // directory permissions.
        $success = $this->mkdir($directory, NULL, TRUE);
        if ($success) {
          return TRUE;
        }
      }
      catch (\Exception $e) {
        // @todo write exception handler to handle Flysystem Exception, and
        // return a proper return code, which is FALSE.
      }
      // If the operation failed, check again if the directory was created
      // by another process/server, only report a failure if not. In this case
      // we still need to ensure the directory is writable.
      if (!$wrapper->directoryExists($directory)) {
        return FALSE;
      }
      return $this->chmod($directory);
    }
    return $this->decoratedService->prepareDirectory($directory, $options);
  }

  /**
   * {@inheritdoc}
   *
   * @todo finish writing, see inline todo comments.
   */
  public function getDestinationFilename($destination, $replace) {
    $basename = $this->basename($destination);
    if (!Unicode::validateUtf8($basename)) {
      throw new FileException(sprintf("Invalid filename '%s'", $basename));
    }
    /** @var \Drupal\Core\StreamWrapper\StreamWrapperInterface $wrapper */
    $wrapper = $this->streamWrapperManager->getViaScheme(StreamWrapperManager::getScheme($destination));
    if ($wrapper instanceof FilesystemAdapter) {
      try {
        if ($wrapper->fileExist($destination)) {
          switch ($replace) {
            case FileSystemInterface::EXISTS_REPLACE:
              // Do nothing here, we want to overwrite the existing file.
              break;

            case FileSystemInterface::EXISTS_RENAME:
              // Hmm, we are using the original Drupal FileSystem::dirname()
              // method here, will it error out?  Do we actually need to call this?
              $directory = $this->dirname($destination);
              $destination = $this->createFilename($basename, $directory);
              break;

            case FileSystemInterface::EXISTS_ERROR:
              // Error reporting handled by calling function.
              return FALSE;
          }
        }
      }
      catch (\Exception $e) {
        // @todo write exception handler to handle Flysystem Exception, and
        // return a proper return code, which is FALSE.
      }
      return $destination;
    }
    return $this->decoratedService->getDestinationFilename($destination, $replace);
  }

  /**
   * {@inheritdoc}
   *
   * @todo finish writing, see inline todo comments.
   * @todo reevaluate against \Drupal\Core\File\FileSystem::createFilename().
   */
  public function createFilename($basename, $directory) {
    $original = $basename;
    // Strip control characters (ASCII value < 32). Though these are allowed in
    // some filesystems, not many applications handle them well.
    $basename = preg_replace('/[\x00-\x1F]/u', '_', $basename);
    if (preg_last_error() !== PREG_NO_ERROR) {
      throw new FileException(sprintf("Invalid filename '%s'", $original));
    }
    if (str_starts_with(PHP_OS, 'WIN')) {
      // These characters are not allowed in Windows filenames.
      $basename = str_replace([':', '*', '?', '"', '<', '>', '|'], '_', $basename);
    }

    // A URI or path may already have a trailing slash or look like "public://".
    if (str_ends_with($directory, '/')) {
      $separator = '';
    }
    else {
      $separator = '/';
    }
    $destination = $directory . $separator . $basename;

    /** @var \Drupal\Core\StreamWrapper\StreamWrapperInterface $wrapper */
    $wrapper = $this->streamWrapperManager->getViaScheme(StreamWrapperManager::getScheme($destination));
    if ($wrapper instanceof FilesystemAdapter) {
      try {
        $exists = $wrapper->fileExists($destination);
        if ($exists) {
          // Destination file already exists, generate an alternative.
          $pos = strrpos($basename, '.');
          if ($pos !== FALSE) {
            $name = substr($basename, 0, $pos);
            $ext = substr($basename, $pos);
          }
          else {
            $name = $basename;
            $ext = '';
          }

          $counter = 0;
          do {
            $destination = $directory . $separator . $name . '_' . $counter++ . $ext;
          } while ($wrapper->fileExists($destination));

        }
      }
      catch (\Exception $e) {
        // @todo write exception handler to handle Flysystem Exception, and
        // return a proper return code, which is FALSE.
      }
      return $destination;
    }
    return $this->decoratedService->createFilename($basename, $directory);
  }

}
